from __future__ import absolute_import

from six import iteritems


class LabelProperty(object):
    """The property with label name."""

    def __init__(self, label, default=None, input_type=None):
        self.label = label
        self.default = default
        self.input_type = input_type
        self.__name__ = None

    def __get__(self, instance, owner):
        if instance is None:
            return self
        if self.__name__ in instance.__dict__:
            return instance.__dict__[self.__name__]
        elif self.default is not None:
            value = self.default()
            instance.__dict__[self.__name__] = value
            return value
        else:
            names = (owner.__name__, self.__name__)
            raise AttributeError('%r object has no attribute %r' % names)

    def __set__(self, instance, value):
        if self.input_type is not None:
            value = self.input_type(value)
        instance.__dict__[self.__name__] = value


_missing = object()


def name_descriptors(cls):
    for name, value in iter_descriptors(cls):
        if getattr(value, '__name__', _missing) is None:
            value.__name__ = name
    return cls


def iter_descriptors(cls):
    for name, value in iteritems(cls.__dict__):
        if name.startswith('__'):
            continue
        if getattr(value, '__get__', None) or getattr(value, '__set__', None):
            yield name, value



from __future__ import absolute_import

from collections import OrderedDict

from .descriptors import LabelProperty, name_descriptors
from .utils import parse_time, parse_peer, parse_vaddr, parse_filesize


@name_descriptors
class Status(object):
    """The OpenVPN status model."""

    client_list = LabelProperty(u'OpenVPN CLIENT LIST', OrderedDict)
    routing_table = LabelProperty(u'ROUTING TABLE', OrderedDict)
    global_stats = LabelProperty(u'GLOBAL STATS')
    updated_at = LabelProperty(u'Updated', input_type=parse_time)


@name_descriptors
class Client(object):
    """The OpenVPN client model."""

    common_name = LabelProperty(u'Common Name')
    real_address = LabelProperty(u'Real Address', input_type=parse_peer)
    bytes_received = LabelProperty(
        u'Bytes Received', input_type=parse_filesize)
    bytes_sent = LabelProperty(u'Bytes Sent', input_type=parse_filesize)
    connected_since = LabelProperty(u'Connected Since', input_type=parse_time)


@name_descriptors
class Routing(object):
    """The OpenVPN routing model."""

    virtual_address = LabelProperty(u'Virtual Address', input_type=parse_vaddr)
    common_name = LabelProperty(u'Common Name')
    real_address = LabelProperty(u'Real Address', input_type=parse_peer)
    last_ref = LabelProperty(u'Last Ref', input_type=parse_time)


@name_descriptors
class GlobalStats(object):
    """The OpenVPN global stats model."""

    max_bcast_mcast_queue_len = LabelProperty(
        u'Max bcast/mcast queue length', input_type=int)



from __future__ import absolute_import

from .parser import LogParser


def parse_status(status_log, encoding='utf-8'):
    """Parses the status log of OpenVPN.

    :param status_log: The content of status log.
    :type status_log: :class:`str`
    :param encoding: Optional. The encoding of status log.
    :type encoding: :class:`str`
    :return: The instance of :class:`.models.Status`
    """
    if isinstance(status_log, bytes):
        status_log = status_log.decode(encoding)
    parser = LogParser.fromstring(status_log)
    return parser.parse()



from __future__ import absolute_import

from six import Iterator, next, text_type

from .models import Status, Client, Routing, GlobalStats
from .descriptors import iter_descriptors


class LogParser(Iterator):
    """The parser for parsing OpenVPN status log.

    This kind of parser is stateful. So the :meth:`LogParser.parse` could be
    called once in the same instance of parser.
    """

    list_separator = u','
    line_separator = u'\n'
    terminator = u'END'

    def __init__(self, lines):
        self.lines = iter(lines)
        self._last_line = None
        self._rollback = False

    def __next__(self):
        if self._rollback:
            self._rollback = False
            return self._last_line
        while True:
            line = next(self.lines).strip()
            if line:
                self._last_line = line
                return line

    @classmethod
    def fromstring(cls, content):
        """Creates a parser from content of log.

        :param str content: The log content.
        :return: The parser instance.
        """
        return cls(content.strip().split(cls.line_separator))

    def rollback(self):
        self._rollback = True

    def expect_line(self, content):
        try:
            line = next(self)
        except StopIteration:
            raise ParsingError('expected %r but got end of input' % content)
        if line != content:
            raise ParsingError('expected %r but got %r' % (content, line))

    def expect_list(self):
        try:
            line = next(self)
        except StopIteration:
            raise ParsingError('expected list but got end of input')
        splited = line.split(self.list_separator)
        if len(splited) == 1:
            raise ParsingError('expected list but got %r' % line)
        return splited

    def expect_tuple(self, name):
        try:
            line = next(self)
        except StopIteration:
            raise ParsingError('expected 2-tuple but got end of input')
        splited = line.split(self.list_separator)
        if len(splited) != 2:
            raise ParsingError('expected 2-tuple but got %r' % line)
        if splited[0] != name:
            raise ParsingError('expected 2-tuple starting with %r' % name)
        return splited[1]

    def parse(self):
        """Parses the status log.

        :raises ParsingError: if syntax error found in the log.
        :return: The :class:`.models.Status` with filled data.
        """
        status = Status()
        self.expect_line(Status.client_list.label)

        status.updated_at = self.expect_tuple(Status.updated_at.label)
        status.client_list.update({
            text_type(c.real_address): c
            for c in self._parse_fields(Client, Status.routing_table.label)})
        status.routing_table.update({
            text_type(r.virtual_address): r
            for r in self._parse_fields(Routing, Status.global_stats.label)})
        status.global_stats = GlobalStats()
        status.global_stats.max_bcast_mcast_queue_len = self.expect_tuple(
            GlobalStats.max_bcast_mcast_queue_len.label)

        self.expect_line(self.terminator)
        return status

    def _parse_fields(self, cls, next_line):
        labels = self.expect_list()
        descriptors = iter_descriptors(cls)
        label_to_name = {
            descriptor.label: name for name, descriptor in descriptors}
        index_to_name = {
            index: label_to_name[label] for index, label in enumerate(labels)}

        while True:
            try:
                values = self.expect_list()
            except ParsingError as list_error:
                try:
                    self.rollback()
                    self.expect_line(next_line)
                except ParsingError as line_error:
                    raise ParsingError(*(list_error.args + line_error.args))
                else:
                    break

            instance = cls()
            for index, value in enumerate(values):
                name = index_to_name[index]
                setattr(instance, name, value)
            yield instance


class ParsingError(Exception):
    pass



from __future__ import absolute_import

import re
import datetime
import collections
import ipaddress

from six import python_2_unicode_compatible as unicode_compatible
from humanize.filesize import naturalsize
from netaddr import EUI, mac_unix


DATETIME_FORMAT_OPENVPN = u'%a %b %d %H:%M:%S %Y'
RE_VIRTUAL_ADDR_MAC = re.compile(
    u'^{0}:{0}:{0}:{0}:{0}:{0}$'.format(u'[a-f0-9]{2}'), re.I)
RE_VIRTUAL_ADDR_NETWORK = re.compile(u'/(\\d{1,3})$')


def parse_time(time):
    """Parses date and time from input string in OpenVPN logging format."""
    if isinstance(time, datetime.datetime):
        return time
    return datetime.datetime.strptime(time, DATETIME_FORMAT_OPENVPN)


def parse_peer(peer):
    if isinstance(peer, PeerAddress):
        return peer
    host, port = peer.rsplit(':', 1)
    return PeerAddress(ipaddress.ip_address(host), int(port))


def parse_vaddr(virtual_addr):
    match = RE_VIRTUAL_ADDR_MAC.search(virtual_addr)
    if match:
        return EUI(virtual_addr, dialect=mac_unix)

    match = RE_VIRTUAL_ADDR_NETWORK.search(virtual_addr)
    if match and 0 < int(match.group(1)) <= 128:
        return ipaddress.ip_network(virtual_addr)

    return ipaddress.ip_address(virtual_addr)


def parse_filesize(size):
    if isinstance(size, FileSize):
        return size
    return FileSize(size)


@unicode_compatible
class PeerAddress(collections.namedtuple('PeerAddress', 'host port')):
    """The address of peer entity.

    :param host: The host address of peer entity.
    :type host: :class:`~ipaddress.IPv4Address` or
                :class:`~ipaddress.IPv6Address`
    :param port: The port of peer entity.
    :type port: :class:`int`
    """

    def __str__(self):
        return '{0}:{1}'.format(self.host, self.port)


@unicode_compatible
class FileSize(int):
    """The size of bytes."""

    def __str__(self):
        return self.humanize()

    def humanize(self, **kwargs):
        """Gets the human-friendly representation of file size.

        :param kwargs: All keyword arguments will be passed to
                       :func:`humanize.filesize.naturalsize`.
        """
        return naturalsize(self, **kwargs)



from setuptools import setup, find_packages


with open('README.rst') as readme:
    next(readme)  # skip badges
    long_description = ''.join(readme).strip()


setup(
    name='openvpn-status',
    version='0.2.0',
    url='https://github.com/tonyseek/openvpn-status',
    author='Jiangge Zhang',
    author_email='tonyseek@gmail.com',
    description='Parse OpenVPN status logs in Python',
    long_description=long_description,
    packages=find_packages(),
    zip_safe=False,
    license='MIT',
    keywords=['openvpn', 'status', 'log'],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Internet :: Log Analysis',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Logging',
        'Topic :: Text Processing',
        'Topic :: Utilities',
    ],
    install_requires=[
        'six>=1.9.0',
        'humanize',
        'netaddr',
    ],
    extras_require={
        ':python_version == "2.7"': [
            'ipaddress',
        ],
    },
    platforms=['Any'],
)
